;; -*- no-byte-compile: t; lexical-binding: nil -*-
(define-package "alert" "20260316.2025"
  "Growl-style notification system for Emacs."
  '((gntp   "0.1")
    (log4e  "0.3.0")
    (cl-lib "0.5"))
  :url "https://github.com/jwiegley/alert"
  :commit "fb8f1fc7f769d7b3ed93b47bd8252f2947c96d9c"
  :revdesc "fb8f1fc7f769"
  :keywords '("notification" "emacs" "message")
  :authors '(("John Wiegley" . "jwiegley@gmail.com"))
  :maintainers '(("John Wiegley" . "jwiegley@gmail.com")))
